# -*- coding: utf-8 -*-

"""
This module contains an implementation of our "heavy-tailed mutation operator,
along with a few utility functions and the classical (1+1) EA.
"""

__version__ = '1.0'

from fastga.operators import *




"""
A few utilitary functions.
"""


def flip(array, i):
    """
    'array' should be an indexable object holding only 0s and 1s.
    This function flips the i-th value in 'array' : if it is 0, it becomes 1, and vice-versa.
    """
    array[i] ^= 1


def wrong_param_error(value, supported_values=None):
    """
    Returns a nicely-formatted error when a given parameter is not of valid value for some function.
    """
    string = "Wrong parameter value '{}'".format(value)
    if supported_values is not None:
        string += (" ; expected values are {{{}}}"
                   .format(", ".join(["{}".format(supported) for supported in supported_values])))
    return ValueError(string + ".")




from numpy import log1p
from numpy.random import rand


__all__ = ["sample", "sample_waiting_time"]


def sample(cdf):
    """
    Samples a random positive bounded integer according to the Cumulative Distribution Function
	given as input (under the form of an array 'cdf', with cdf[i + 1] >= cdf[i]
    and cdf[-1] = 1. Note that these verifications are not performed by the function).
    
    The resulting integer X has the property that
        Pr(X <= i + 1) = cdf[i], for i in [0, ..., n - 1] :
    
    Pr(X = 1) = cdf[0]
    Pr(X = 2) = cdf[1] - cdf[0]
    ...
    Pr(X = n) = cdf[n - 1] - cdf[n - 2] - ... - cdf[0]

    """
    n = len(cdf)
    x = rand()
    if x < cdf[0]:
        return 1
    low = 0
    high = n - 1
    while high - low > 1:
        mid = (low + high) >> 1  # same as (low + high) // 2, but hopefully (mildly) faster
        if x > cdf[mid]:
            low = mid
        else:
            high = mid
    return high + 1


def sample_waiting_time(p):
    """
    Samples the number of independent tries until a given random event happens.
    'p' should be the probability of the event happening in one try.
    """
    x = rand()
    # We use log1p(z) to compute precisely log(1 + z) :
    return 1 + int(log1p(- x) // log1p(- p))




#!/usr/bin/env python
# -*- coding: utf-8 -*-
 
from setuptools import setup, find_packages
 
__version__ = '1.0.0'

min_numpy_ver = '1.7.0'

setup(
 
    name='fastga',
    version=__version__,

    author='The FastGA team',

    license="Unlicense",
 
    packages=find_packages(),
 
    description="Implementation of the so-called Heavy-Tailed Mutation Operator",
 
    long_description=open('README.rst').read(),
 
    install_requires=["numpy >= {}".format(min_numpy_ver)],
 
    # Active la prise en compte du fichier MANIFEST.in
    include_package_data=True,
 
    # Une url qui pointe vers la page officielle de votre lib
    url='https://github.com/FastGA/fast-genetic-algorithms',
 
    # Il est d'usage de mettre quelques metadata à propos de sa lib
    # Pour que les robots puissent facilement la classer.
    # La liste des marqueurs autorisées est longue:
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers.
    #
    # Il n'y a pas vraiment de règle pour le contenu. Chacun fait un peu
    # comme il le sent. Il y en a qui ne mettent rien.
    classifiers=[
        "Programming Language :: Python",
        "Operating System :: OS Independent",
        "License :: Public Domain",
        "Programming Language :: Python :: 3.5",
        "Topic :: Scientific/Engineering :: Artificial Intelligence"
    ],
)



"""
Implementation of the different mutation operators.

BaseMutationOperator is an abstract class that sports the basic methods to perform the bitwise mutation of a bit string,
OnePlusOneMutationOperator is an implementation of the classical (1+1) EA,
FastMutationOperator is an implementation of our operator.

The inheritance scheme is

    BaseMutationOperator
        |- OnePlusOneMutationOperator
        |- FastMutationOperator

and a sample usage is :

    oneplusone = OnePlusOneMutationOperator(n=100)
    fast = FastMutationOperator(n=100, beta=1.5)

"""

from numpy import linspace
from fastga.utils import flip, wrong_param_error
from fastga.sampling import sample, sample_waiting_time


__all__ = ["BaseMutationOperator", "OnePlusOneMutationOperator", "FastMutationOperator"]


class BaseMutationOperator:
    """
    This abstract class holds the common methods that one would expect in a mutation operator.
    It should not be used directly.
    """

    def __init__(self, n): 
        if not isinstance(n, int) or n < 1:
            raise ValueError("Parameter 'n' must be a positive integer. Received {}.".format(n))
        self.n = n

    def get_mutation_rate(self):
        raise NotImplementedError() # this should be implemented in subclasses

    def _mutate(self, bit_string):
        """
        Private 'mutate' method, not to be called externally. Performs an inplace mutation of
        the input bit string. A naive implementation would be to go through the n bits of the
        bit string and flip them with probability given by the mutation rate.
        Alternatively, one can sample the indexes of bits to flip, and flip exactly those. This
        is what we do here.
        """
        n = self.n
        mutation_rate = self.get_mutation_rate()
        index = sample_waiting_time(mutation_rate)
        while index < n:
            flip(bit_string, index)
            index += sample_waiting_time(mutation_rate)

    def mutate(self, bit_string, inplace=False):
        """
        Public 'mutate' method ; can perform the mutation inplace or return a (new)
        mutated bit string.
        """
        n = self.n
        bit_string_len = len(bit_string)
        if not bit_string_len == n:
            raise ValueError("Wrong length for input bit string : expected {}, got {}."
                    .format(n, bit_string_len))
            if not inplace:
                bit_string = bit_string.copy()
        self._mutate(bit_string)
        if not inplace:
            return bit_string


class OnePlusOneMutationOperator(BaseMutationOperator):

    def __init__(self, n):
        super().__init__(n)

    def get_mutation_rate(self):
        return 1 / self.n


class FastMutationOperator(BaseMutationOperator):

    def __init__(self, n, beta):
        super().__init__(n)
        if beta <= 1:
            raise ValueError("Wrong value for parameter beta : received {}, must be a float > 1."
                    .format(beta))
            self.beta = beta
        # We pre-compute the power law cumulative distribution function with the given parameter
        # beta. It will allow us to sample random integers according to this distribution.
        power_law = (linspace(1, n//2, n//2) ** (- beta)).cumsum()
        power_law /= power_law[-1]
        self.power_law = power_law

    def get_mutation_rate(self):
        r = sample(self.power_law)
        return r / self.n


if __name__ == '__main__':
    fast = FastMutationOperator(10, 1.5)
    bit_string = [0] * 10
    print(bit_string)
    for i in range(10):
        fast.mutate(bit_string, inplace=True)
        print(bit_string)

